# -*- coding: iso-8859-1 -*-
# pylint: disable-msg=W0613,W0706
# pylint-version = 0.7.0
#
# Copyright 2004-2005 Andr Malo or his licensors, as applicable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test runner for the svnmailer test suite
"""
__author__    = "Andr Malo"
__docformat__ = "restructuredtext en"
__all__       = ['Tester', 'TestCase', 'loadModule']

# global imports
import fnmatch, locale, os, re, sys, unittest

# global variables
_DIRRE = re.compile(r'[a-zA-Z_][a-zA-Z0-9_]*$')
_FILERE = re.compile(r'[a-zA-Z_][a-zA-Z0-9_]*\.py$')

# exceptions
class SkipException(Exception):
    """ Test should be skipped """
    pass


def loadModule(name):
    """ loads a module """
    mod = __import__(name)
    components = name.split('.')
    for comp in components[1:]:
        mod = getattr(mod, comp)

    return mod


def findModules(packages, filenames = True):
    """ Find all modules in packages (includes the packages themselves) """
    packs = {}
    for package in packages:
        try:
            thismod = loadModule(package)
        except ImportError:
            break # was not a package

        for prefix in thismod.__path__:
            for path, dirs, files in os.walk(prefix, True):
                if '__init__.py' in files or '__init__.pyc' in files:
                    mods = [name for name in files if _FILERE.match(name)]
                    if filenames:
                        mods = [os.path.join(path, name) for name in mods]
                    else:
                        mods = ["%s.%s" % (package, name[:name.find('.')])
                            for name in mods
                            if not name.startswith('__init__.')
                        ]
                    packs.setdefault(package, []).extend(mods)
                    packs.update(findModules([
                        "%s.%s" % (package, name) for name in dirs
                        if _DIRRE.match(name)
                    ], filenames))
                dirs[:] = []

    return packs


class TestCase(unittest.TestCase):
    """ Modified TestCase class, which allows for skipped tests """

    def skip(self, reason = None):
        """ Skip this test """
        raise SkipException(reason)


class TestResult(unittest.TestResult):
    """ Modified TestResult class, which allows for skipped tests """

    def __init__(self):
        """ Initialization """
        super(TestResult, self).__init__()
        self.testsSkipped = 0


    def addError(self, test, err):
        """ Redirects SkipExceptions to L{addSkip} """
        # HACK ALERT. Maybe we need to rewrite our own complete unittest impl.
        if str(err[0]).endswith('unittest.SkipException'):
            self.testsSkipped += 1
            self.addSkip(test, err[1])
            return False
        else:           
            super(TestResult, self).addError(test, err)
            return True


    def addSkip(self, test, reason):
        """ Called when a test was skipped """
        pass


class TextTestResult(TestResult):
    """ Prints test details on the stream """
    separator1 = '=' * 70 + "\n"
    separator2 = '-' * 70 + "\n"

    def __init__(self, outstream, verbosity):
        """ Initialization """
        super(TextTestResult, self).__init__()
        self._out = outstream
        self._verbosity = verbosity


    def getDescription(self, test):
        """ Returns the test description """
        return test.shortDescription() or str(test)


    def startTest(self, test):
        """ Called when the test is about to start """
        super(TextTestResult, self).startTest(test)
        if self._verbosity >= 2:
            self._out.write(self.getDescription(test))
            self._out.write(" ... ")


    def addSuccess(self, test):
        """ Called when the test was successful """
        super(TextTestResult, self).addSuccess(test)
        if self._verbosity:
            self._out.write([".", "ok\n"][self._verbosity >= 2])


    def addError(self, test, err):
        """ Called when the test failed with an unexpected exception """
        if super(TextTestResult, self).addError(test, err):
            if self._verbosity:
                self._out.write(["E", "ERROR\n"][self._verbosity >= 2])


    def addFailure(self, test, err):
        """ Called when the test failed """
        super(TextTestResult, self).addFailure(test, err)
        if self._verbosity:
            self._out.write(["F", "FAIL\n"][self._verbosity >= 2])


    def addSkip(self, test, reason):
        """ Called when the test was skipped """
        super(TextTestResult, self).addSkip(test, reason)
        skipped = "SKIPPED"
        if reason is not None and self._verbosity >= 2:
            skipped = "%s (%s)" % (skipped, reason)
        if self._verbosity:
            self._out.write(["S", "%s\n" % skipped][self._verbosity >= 2])


    def printErrors(self):
        """ Prints all error/failure lists """
        if self._verbosity:
            self._out.write("\n")
        self._printErrorList('ERROR', self.errors)
        self._printErrorList('FAIL', self.failures)


    def _printErrorList(self, flavour, errors):
        """ Prints one error/failure list """
        for test, err in errors:
            self._out.write(self.separator1)
            self._out.write("%s: %s\n" % (flavour, self.getDescription(test)))
            self._out.write(self.separator2)
            self._out.write("%s\n" % err)


class TextTestRunner(object):
    """ Runs a suite and prints the progression on the screen """

    def __init__(self, outstream, verbosity):
        """ Initialization """
        self._out = outstream
        self._verbosity = verbosity


    def run(self, suite):
        """ Runs the suite """
        result = self._createResult()
        suite(result)
        if result.failures or result.errors:
            result.printErrors()

        if self._verbosity:
            if self._verbosity == 1:
                self._out.write("\n")
            self._printSummary(result)

        return result


    def _printSummary(self, result):
        """ Prints the summary """
        if not result.wasSuccessful() or self._verbosity > 1:
            self._out.write(result.separator2)
        self._out.write("%s - %d test%s total" % (
            [["FAILED", "ERROR"][bool(result.errors)],
                "OK"][result.wasSuccessful()],
            result.testsRun,
            ["s", ""][result.testsRun == 1],
        ))
        if not result.wasSuccessful() or result.testsSkipped:
            self._out.write(" (")
            parts = []
            if result.testsSkipped:
                parts.append("%d skipped" % result.testsSkipped)
            if result.failures:
                parts.append("%d failed" % len(result.failures))
            if result.errors:
                parts.append("%d errored" % len(result.errors))
            self._out.write(", ".join(parts))
            self._out.write(")")

        self._out.write("\n")


    def _createResult(self):
        """ Returns the result object """
        return TextTestResult(self._out, self._verbosity)


class UnicodeStreamDecorator(object):
    """ Unicode writer """

    def __init__(self, stream):
        """ Initialization

            :param stream: The stream to decorate
            :type stream: ``file``
        """
        self._stream = stream
        self._encoding = self._getOutputEnc(stream)


    def __getattr__(self, attr):
        """ Returns not decorated attributes

            :param attr: The attribute name
            :type attr: ``str``

            :return: The attribute
            :rtype: any

            :exception AttributeError: The attribute doesn't exist
        """
        return getattr(self._stream, attr)


    def write(self, towrite):
        """ Write a string to the stream """
        if isinstance(towrite, unicode):
            towrite = towrite.encode(self._encoding, 'replace')

        self._stream.write(towrite)
        self._stream.flush()


    def _getOutputEnc(self, stream):
        """ Returns the "proper" output encoding

            If the output goes to a terminal, the method tries to get
            the current locale encoding. UTF-8 is default and fallback
            if anything fails.

            :param stream: The file object written to
            :type stream: ``file``

            :return: The chosen encoding
            :rtype: ``str``
        """
        enc = "utf-8"
        try:
            isatty = os.isatty(stream.fileno())
        except AttributeError:
            isatty = False

        if isatty:
            enc = locale.getpreferredencoding() or enc

        return enc


class Tester(object):
    """ Tester class """
    packages = ('svnmailer',)
    testpackage = 'svnmailer_test'
    suites = (
        ('unit', "Unit Tests"),
        ('integration', "Integration Tests"),
    )

    def __init__(self, cover = 0, outstream = None, verbose = 2, suites = ()):
        """ Initialization """
        if outstream is None:
            outstream = sys.stdout

        self._h_begin, self._h_end = self._getTerminalEscapes(outstream)
        self._out = UnicodeStreamDecorator(outstream)
        self._cover = cover
        self._suites = suites
        self._verbosity = verbose

        if cover:
            self._coverage = loadModule("%s.coverage" % self.testpackage)
            self._coverage.erase() # cleanup


    def cli(cls, argv):
        """ Alternative constructor (classmethod)

            :return: Tester class
            :rtype: ``Tester``
        """
        argv = list(argv)
        cover = 0
        verbose = 1
        opt = argv and argv[0]
        while opt in ['-q', '-v', '-c', '-r']:
            argv.pop(0)
            if opt in ['-c', '-r']:
                cover = 1 + (opt == '-c')
            elif opt == '-v':
                verbose = 2
            elif opt == '-q':
                verbose = 0
            opt = argv and argv[0]

        return cls(cover = cover, suites = argv, verbose = verbose)

    cli = classmethod(cli)


    def _getTerminalEscapes(self, outstream):
        """ Returns the terminal escapes for bright and normal output

            :return: ('bright', 'normal')
            :rtype: ``tuple``
        """
        try:
            isatty = os.isatty(outstream.fileno())
        except AttributeError:
            isatty = False

        if isatty and sys.platform != 'win32':
            bright, normal = ("\033[1m", "\033[0m")
        else:
            bright, normal = ("", "")

        return bright, normal


    def getSuites(self):
        """ Returns the test suites

            :return: The test suite list; each list item is a tuple, where
                     the first item is the test module object and the second
                     item the suite object.
            :rtype: ``generator``
        """
        old_group = None
        for group, grouptitle in self.suites:
            packages = findModules(['%s.%s' % (self.testpackage, group)
                ], False).items()
            packages.sort()

            for package, mods in packages:
                mods = mods + [package]
                mods.sort()
                pats = self._suites and self._suites or ('',)
                for mod in mods:
                    cases = []
                    mobj = loadModule(mod)
                    if not hasattr(mobj, 'getSuite'):
                        continue

                    for testcase in mobj.getSuite():
                        name = "%s.%s" % (mod, testcase)
                        for pattern in pats:
                            if fnmatch.fnmatch(name, "*%s" % pattern):
                                cases.append(testcase)
                                break
                    if cases:
                        if old_group != group:
                            self.pregroup(group, grouptitle)
                            old_group = group
                        yield mod, unittest.TestLoader().loadTestsFromNames(
                            cases, mobj
                        )


    def _echogroup(self, grouptitle):
        """ Prints the groupname in a fancy form """
        if self._verbosity:
            self._out.write('\n')
        self._center(" %s " % grouptitle)
        if self._verbosity:
            self._out.write('\n')

        
    def _echosuite(self, suite):
        """ Prints the suitename in a fancy form """
        self._out.write(
            "%(begin)s>>> Running testsuite %(suite)s.*%(end)s\n" % {
                'suite': suite[0],
                'begin': self._h_begin,
                'end':   self._h_end,
            }
        )


    def pregroup(self, group, grouptitle):
        """ Runs before each new group """
        self._echogroup(grouptitle)


    def prerun(self, suite):
        """ Runs before every suite """
        self._echosuite(suite)
        if self._cover:
            self._coverage.start()


    def postrun(self, suite):
        """ Runs after every suite """
        if self._cover:
            self._coverage.stop()


    def run(self):
        """ Runs the tests """
        try:
            return self._dorun()
        except KeyboardInterrupt:
            self._out.write("\n-> interrupted\n")
            return 1


    def _dorun(self):
        """ Really runs the tests """
        errors = 0
        success = 0
        skipped = 0
        rerun = []

        for suite in self.getSuites():
            self.prerun(suite)

            result = TextTestRunner(
                outstream = self._out,
                verbosity = self._verbosity
            ).run(suite[1])

            self.postrun(suite)

            if result.wasSuccessful():
                success += result.testsRun
            else:
                this_errors = len(result.errors) + len(result.failures)
                errors += this_errors
                success += result.testsRun - this_errors
                rerun.append(suite[0])
            skipped += result.testsSkipped
    
            if self._verbosity:
                self._out.write('\n')

        self.summary(success, skipped, errors, rerun)
        if self._cover:
            self.coverageReport()

        if errors:
            return 1

        return 0


    def _center(self, string, width = 70):
        """ prints a centered string """
        self._out.write(
            string.replace(' ', '\0').center(width
                ).replace(' ', '=').replace('\0', ' ') + "\n"
        )


    def summary(self, success, skipped, errors, rerun):
        """ Prints the summary """
        if not self._verbosity:
            self._out.write("\n")
        self._center(" SUMMARY ")
        self._out.write("Tests Succeeded: %d\n" % success)
        self._out.write("Tests Skipped  : %d\n" % skipped)
        self._out.write("Tests Failed   : %d\n" % errors)
        if rerun:
            self._out.write("  %s\n" % '\n  '.join([
                suite.replace("%s." % self.testpackage, '', 1)
                for suite in rerun
            ]))


    def coverageReport(self):
        """ Prints the coverage report """
        self._out.write("\n")
        self._center(" COVERAGE REPORT ")
        packages = findModules(self.packages).items()
        packages.sort()

        for package, mods in packages:
            self._out.write("%s>>> package %s%s\n" %
                (self._h_begin, package, self._h_end)
            )
            mods.sort()
            old_stdout = sys.stdout
            try:
                sys.stdout = self._out
                self._coverage.report(mods, False)
                if self._cover == 2:
                    self._coverage.the_coverage.annotate(mods)
            finally:
                sys.stdout = old_stdout

            self._out.write("\n")


if __name__ == '__main__':
    Tester.cli(sys.argv[1:]).run()
