# -*- coding: iso-8859-1 -*-
# pylint: disable-msg=C0103,E0201,W0131,W0613,E0213,W0232,W0201
# pylint-version = 0.7.0
#
# Copyright 2005 Andr Malo or his licensors, as applicable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Package svnmailer_test.unit.browser

Testcases for svnmailer.browser

:var _tests: The test classes
:type _tests: ``list``
"""
__author__    = "Andr Malo"
__docformat__ = "restructuredtext en"
__all__       = ['getSuite']

from svnmailer_test import base
_tests = []


#### Test Bases                                     ####
########################################################
class TestBase(base.TestBase):
    """ Base class for browser tests """
    LOAD_MODULES = (('_browser', 'svnmailer.browser'),)

    def getInitialMocks(self):
        """ Returns the initial mocks """
        return [
            self._mock_svnmailer_util(),
        ]


#### Test Cases                                     ####
########################################################
class _init_ErrorTestCase(TestBase):
    """ svnmailer.browser.Error """

    def test00general(self):
        """ Error is a public Exception """
        expected = Exception
        result = self._browser.Error

        self.failUnless(issubclass(result, expected),
            "Is not an %r, but %r" % (expected, result))
        self.failUnless(result.__name__ in self.public(self._browser),
            "Is not public")

_tests.append(_init_ErrorTestCase)


class _init_InvalidGeneratorErrorTestCase(TestBase):
    """ svnmailer.browser.InvalidBaseUrlError """

    def test00general(self):
        """ InvalidGeneratorError is a public Error """
        expected = self._browser.Error
        result = self._browser.InvalidGeneratorError

        self.failUnless(issubclass(result, expected),
            "Is not an %r, but %r" % (expected, result))
        self.failUnless(result.__name__ in self.public(self._browser),
            "Is not public")

_tests.append(_init_InvalidGeneratorErrorTestCase)


class _init_ManagerTestCase(TestBase):
    """ svnmailer.browser.Manager """
    PREFIX = "Manager"

    def test00general(self):
        """ Manager is a public util.Singleton """
        expected = self.mock['svnmailer.util'].Singleton
        result = self._browser.Manager

        self.failUnless(issubclass(result, expected),
            "Is not an %r, but %r" % (expected, result))
        self.failUnless(result.__name__ in self.public(self._browser),
            "Is not public")

    def test01initial(self):
        """ ._GENERATORS is properly filled initially """
        self.mock.result = [self._browser.Manager._GENERATORS]

        self.expect({
            u"viewcvs" : 'svnmailer.browser.viewcvs.Generator',
            u"websvn"  : 'svnmailer.browser.websvn.Generator',
            u"svn::web": 'svnmailer.browser.svnweb.Generator',
            u"chora"   : 'svnmailer.browser.chora.Generator',
            u"trac"    : 'svnmailer.browser.trac.Generator',
            u"generic" : 'svnmailer.browser.generic.Generator',
        })

    def test02registerGenerator(self):
        """ .registerGenerator registers the new class name """
        inst = self._browser.Manager()
        inst.registerGenerator(u"SomeType", 'some.class')
        self.mock.result = [self._browser.Manager._GENERATORS[u'sometype']]

        self.expect('some.class')

    def _mockForSelect(self, config, error = None):
        """ Mocks for select tests """
        class GenClass(object):
            __mocked__ = 'Generator'
            def __init__(this, base_url, config):
                this._mocked = this.__mocked__
                self.mock.result.extend(['GenClass', base_url, config])
            def hasTemplates(this):
                return self.mock.error != 'notemplates'

        class ViewcvsClass(GenClass):
            __mocked__ = 'ViewcvsGenerator'
        class GenericClass(GenClass):
            __mocked__ = 'GenericGenerator'

        class MyManager(self._browser.Manager):
            def _parseBrowserBase(this, base_config):
                self.mock.result.extend(['_parseBrowserBase', base_config])
                return ('btype', 'burl')
            def _load(this, name):
                if error == 'load.key':
                    raise KeyError()
                elif error == 'load.import':
                    raise ImportError()
                self.mock.result.extend(['_load', name])
                if name == u'viewcvs':
                    return ViewcvsClass
                elif name == u'generic':
                    return GenericClass
                return GenClass

        self._mymanager = MyManager

    def test03select(self):
        """ .select returns the proper generator """
        config = base.Store(browser_base_url = 'abc')
        self._mockForSelect(config)
        self.mock.result.append(self._mymanager().select(config)._mocked)

        self.expect(
            '_parseBrowserBase', 'abc',
            '_load', 'btype',
            'GenClass', 'burl', config,
            'Generator'
        )

    def test04select(self):
        """ .select raises InvalidGeneratorError """
        config = base.Store(browser_base_url = 'abc')
        self._mockForSelect(config, 'load.key')
        self.failUnlessRaises(self._browser.InvalidGeneratorError,
            self._mymanager().select, config)

    def test05select(self):
        """ .select passes ImportError """
        config = base.Store(browser_base_url = 'abc')
        self._mockForSelect(config, 'load.import')
        self.failUnlessRaises(ImportError, self._mymanager().select, config)

    def test06select(self):
        """ .select falls back to viewcvs_base_url """
        config = base.Store(browser_base_url = None, viewcvs_base_url = 'x')
        self._mockForSelect(config)
        self.mock.result.append(self._mymanager().select(config)._mocked)

        self.expect(
            '_load', u'viewcvs',
            'GenClass', 'x', config,
            'ViewcvsGenerator'
        )

    def test07select(self):
        """ .select falls back to generic templates """
        config = base.Store(browser_base_url = None, viewcvs_base_url = None)
        self._mockForSelect(config)
        self.mock.result.append(self._mymanager().select(config)._mocked)

        self.expect(
            '_load', u'generic',
            'GenClass', None, config,
            'GenericGenerator'
        )

    def test08select(self):
        """ .select drops non-template generators """
        config = base.Store(browser_base_url = 'abc')
        self.mock.error = 'notemplates'
        self._mockForSelect(config)
        self.mock.result.append(self._mymanager().select(config))

        self.expect(
            '_parseBrowserBase', 'abc',
            '_load', 'btype',
            'GenClass', 'burl', config,
            None
        )

    def test09parseBrowserBase(self):
        """ ._parseBrowserBase parses 2 tokens """
        self.mock.result = [
            self._browser.Manager()._parseBrowserBase('A \tB c')
        ]
        self.expect(('a', 'B c'))

    def test10parseBrowserBase(self):
        """ ._parseBrowserBase parses 1 token """
        self.mock.result = [self._browser.Manager()._parseBrowserBase('a ')]
        self.expect(('a', None))

    def test11parseBrowserBase(self):
        """ ._parseBrowserBase ignores empty input """
        for inp in ('', None, '  '):
            self.mock.result = [self._browser.Manager()._parseBrowserBase(inp)]
            self.expect((None, None))

    def test12load(self):
        """ ._load returns the dotted object """
        inst = self._browser.Manager()
        inst.registerGenerator(u"some", "some.class")
        self.mock.result.append(inst._load(u'some')._mocked)

        self.expect(
            'util.loadDotted', 'some.class',
            'SomeClass'
        )

    def test13load(self):
        """ ._load expects a qualified name """
        inst = self._browser.Manager()
        inst.registerGenerator(u"some", "class")

        self.failUnlessRaises(ImportError, inst._load, u'some')

    def test14load(self):
        """ ._load passes ImportError """
        inst = self._browser.Manager()
        inst.registerGenerator(u"some", "foo.class")

        self.failUnlessRaises(ImportError, inst._load, u'some')

_tests.append(_init_ManagerTestCase)


class _init_utilTestCase(TestBase):
    """ svnmailer.browser.util """
    
    def test00delete(self):
        """ The util module import is not exposed to the public """
        self.failIf(hasattr(self._browser, 'util'), "util attribute found")

_tests.append(_init_utilTestCase)


#### Glue code                                      ####
########################################################
__all__.extend(_tests)

def getSuite():
    """ Returns the test suite for this module """
    return [_test.__name__ for _test in _tests]
