# -*- coding: iso-8859-1 -*-
# pylint: disable-msg=C0103,E0201,W0131,E0213,W0232,F0203,W0613,R0904
# pylint-version = 0.7.0
#
# Copyright 2005 Andr Malo or his licensors, as applicable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Testcases for svnmailer.settings._typedstruct

:var _tests: The test classes
:type _tests: ``list``
"""
__author__    = "Andr Malo"
__docformat__ = "restructuredtext en"
__all__       = ['getSuite']

from svnmailer_test import base
_tests = []


#### Test Bases                                     ####
########################################################
class TestBase(base.TestBase):
    """ Base class for settings._typedstruct tests """
    LOAD_MODULES = (('_struct', 'svnmailer.settings._typedstruct'),)

    def getInitialMocks(self):
        """ Returns the initially mocked modules """
        return [
            self._mock_svnmailer_util(),
        ]


#### Test Cases                                     ####
########################################################
class MemberTestCase(TestBase):
    """ svnmailer.settings._typedstruct.Member """
    PREFIX = "Member"

    def setUp(self):
        """ Extends test environment """
        super(MemberTestCase, self).setUp()
        class MyMember(self._struct.Member):
            def __init__(this):
                pass

        self._mymember = MyMember

    def test00general(self):
        """ Member is a public object """
        expected = object
        result = self._struct.Member

        self.failUnless(issubclass(result, expected),
            "Is not an %r, but %r" % (expected, result))
        self.failUnless(result.__name__ in self.public(self._struct),
            "Is not public")

    def test01init(self):
        """ .__init__ initializes the attributes and calls .init() """
        class MyMember(self._struct.Member):
            def init(this):
                self.mock.result.extend([
                    this.mapper, this.arg, this.param, this.instance
                ])
        MyMember('*mapper*', '*arg*', '*param*')

        self.expect('*mapper*', '*arg*', '*param*', None)

    def test02init(self):
        """ .init is implemented """
        self._mymember().init()

    def test03premap(self):
        """ .premap passes the input """
        self.mock.result.append(self._mymember().premap(self))

        self.expect(self)

    def test04transform(self):
        """ .transform passes the input """
        self.mock.result.append(self._mymember().transform(self))

        self.expect(self)

    def test05substitute(self):
        """ .substitute passes the input """
        self.mock.result.append(
            self._mymember().substitute(self, {'x': 'y'})
        )

        self.expect(self)

    def test06postmap(self):
        """ .postmap passes the input """
        self.mock.result.append(self._mymember().postmap(self))

        self.expect(self)

_tests.append(MemberTestCase)


class MetaClassTestCase(TestBase):
    """ svnmailer.settings._typedstruct.MetaClass """
    PREFIX = "MetaClass"

    def test00general(self):
        """ MetaClass is a public type """
        expected = type
        result = self._struct.MetaClass

        self.failUnless(issubclass(result, expected),
            "Is not an %r, but %r" % (expected, result))
        self.failUnless(result.__name__ in self.public(self._struct),
            "Is not public")

    def test01new(self):
        """ .__new__ removes mcs.__dict__ and adds __slots__ """
        class Foo(object):
            __metaclass__ = self._struct.MetaClass

        inst = Foo()
        self.failIf(hasattr(inst, '__dict__'), "Has dict")
        self.failUnless(hasattr(inst, '__slots__'), "No slots")
        self.mock.result.append(inst.__slots__)

        self.expect(())

    def test02new(self):
        """ .__new__ takes over existing slots """
        class Foo(object):
            __metaclass__ = self._struct.MetaClass
            __slots__ = ['a', 'b']

        inst = Foo()
        self.failIf(hasattr(inst, '__dict__'), "Has dict")
        self.failUnless(hasattr(inst, '__slots__'), "No slots")
        self.mock.result.append(inst.__slots__)

        self.expect(('a', 'b'))

_tests.append(MetaClassTestCase)


class StructTestCase(TestBase):
    """ svnmailer.settings._typedstruct.Struct """
    PREFIX = "Struct"

    def test00general(self):
        """ Struct is a public object """
        expected = object
        result = self._struct.Struct

        self.failUnless(issubclass(result, expected),
            "Is not an %r, but %r" % (expected, result))
        self.failUnless(result.__name__ in self.public(self._struct),
            "Is not public")

    def test01initial(self):
        """ .__metaclass__ is set properly """
        self.mock.result.append(self._struct.Struct.__metaclass__)

        self.expect(self._struct.MetaClass)

    def test02init(self):
        """ .__init__ sets the given attributes properly """
        class Foo(self._struct.Struct):
            __slots__ = ['a', 'b']
        inst = Foo(a = 1, b = 2)
        self.mock.result.extend([inst.a, inst.b])

        self.expect(1, 2)

    def test03init(self):
        """ .__init__ raises AttributeError """
        class Foo(self._struct.Struct):
            __slots__ = ['a', 'b']

        self.failUnlessRaises(AttributeError, Foo, a = 1, b = 2, c =3)

_tests.append(StructTestCase)


class StructCreatorTestCase(TestBase):
    """ svnmailer.settings._typedstruct.StructCreator """
    PREFIX = "StructCreator"

    def setUp(self):
        """ Extends the test environment """
        super(StructCreatorTestCase, self).setUp()
        class SomeClass(object):
            pass

        def instancemethod(func, inst, cls):
            self.mock.result.extend(['instancemethod', inst, cls])
            return "instancemethod(%s)" % func

        def member1(mapper, arg, param):
            return "member1(%s, %s, %s)" % (mapper, arg, param)

        def member2(mapper, arg, param):
            return "member2(%s, %s, %s)" % (mapper, arg, param)

        class MyCreator0(self._struct.StructCreator):
            _DEFAULTMEMBER = '*member*'
            _DEFAULTSTRUCT = '*struct*'
            def _prepareMembers(this, members, typemap):
                if self.mock.error == 'prepare':
                    raise AssertionError()
                self.mock.result.append('_prepareMembers')
                new = {}
                for key, value in members.items():
                    new[(key,)] = value
                return new, tuple(members.keys())

        class MyCreator(self._struct.StructCreator):
            _BASESTRUCT    = base.Store
            def __init__(this):
                this._cls = SomeClass
                this._aliases = {}
                this._members = {
                    ('foo',): (member1, None), ('bar',): (member2, 1)}
                this._names = ('foo', 'bar')
                this._eqignore = None

        class MyCreator2(MyCreator):
            def _createDescriptor(this, name, private, member):
                return "descriptor(%s, %s, %s)" % (name, private, member)

        self._struct.instancemethod = instancemethod
        self._mycreator0 = MyCreator0
        self._mycreator  = MyCreator
        self._mycreator2 = MyCreator2

    def test00general(self):
        """ StructCreator is a public object """
        expected = object
        result = self._struct.StructCreator

        self.failUnless(issubclass(result, expected),
            "Is not an %r, but %r" % (expected, result))
        self.failUnless(result.__name__ in self.public(self._struct),
            "Is not public")

    def test01initial(self):
        """ StructCreator's class attributes are properly filled initially """
        cls = self._struct.StructCreator
        self.mock.result.extend([
            cls._BASESTRUCT, cls._DEFAULTSTRUCT, cls._DEFAULTMEMBER
        ])

        self.expect(
            self._struct.Struct, self._struct.Struct, self._struct.Member
        )

    def test02init(self):
        """ .__init__ initializes from members """
        inst = self._mycreator0({'a': 'b', 'c': 'd'})
        self.mock.result.extend([
            inst._cls, inst._members, inst._aliases, inst._names,
            inst._eqignore
        ])
        
        self.expect(
            '_prepareMembers', '*struct*',
            {('a',): 'b', ('c',): 'd'}, {}, ('a', 'c'), None
        )

    def test03init(self):
        """ .__init__ passes AssertionError """
        self.mock.error = 'prepare'
        self.failUnlessRaises(AssertionError,
            self._mycreator0, {'a': 'b', 'c': 'd'})

    def test04init(self):
        """ .__init__ initializes ._cls from cls """
        class Foo(object):
            pass
        inst = self._mycreator0({}, cls = Foo)
        self.mock.result.append(inst._cls)

        self.expect('_prepareMembers', Foo)

    def test05init(self):
        """ .__init__ initializes ._aliases from aliases """
        inst = self._mycreator0({}, aliases = {'foo': 'bar'})
        self.mock.result.append(inst._aliases)

        self.expect('_prepareMembers', {'foo': 'bar'})

    def test06init(self):
        """ .__init__ initializes ._eqignore from eqignore """
        inst = self._mycreator0({}, eqignore = [])
        self.mock.result.append(inst._eqignore)
        inst = self._mycreator0({}, eqignore = ['a', 'b'])
        self.mock.result.append(inst._eqignore)

        self.expect(
            '_prepareMembers', {},
            '_prepareMembers', {'a': None, 'b': None}
        )

    def test07create(self):
        """ .create returns a new instance of ._cls """
        class Foo(object):
            def __new__(cls, **initkw):
                self.mock.result.extend(['Foo', initkw])
                initkw['__cls'] = True
                return initkw

        class MyCreator(self._mycreator):
            def _createPrivate(this):
                self.mock.result.extend(['_createPrivate'])
                return '*private*'
            def _generateMembers(this, private, maps, arg):
                self.mock.result.extend(['_generateMembers',
                    private, maps, arg])
                return {'a': 'adesc', 'b': 'bdesc'}
            def _createMetaClass(this, name, bases, cdict):
                self.mock.result.extend(['_createMetaClass',
                    name, bases, cdict])
                return Foo
            def _generateSpecials(this, private):
                self.mock.result.extend(['_generateSpecials', private])
                return {'special': 'somefunc'}
        inst = MyCreator()
        cls = inst._cls
        self.mock.result.append(inst.create())
        inst._aliases = {'x': 'map1'}
        self.mock.result.append(inst.create(
            maps = {'map1': 'mapper', 'x': 'mapper2'}, arg = '*arg*',
            initkw = {'init': 'kw'}
        ))

        self.expect(
            '_createPrivate',
            '_generateMembers', '*private*', {}, None,
            '_createMetaClass', 'SomeClass', (cls,),
                {'a': 'adesc', 
                '__module__': 'svnmailer_test.unit.settings._typedstruct',
                'b': 'bdesc', '__slots__': ('foo', 'bar')},
            '_generateSpecials', '*private*',
            'instancemethod', None, Foo,
            'Foo', {'__cls': True},
            {'__cls': True},
            # second
            '_createPrivate',
            '_generateMembers', '*private*', {'map1': 'mapper2'}, '*arg*',
            '_createMetaClass', 'SomeClass', (cls,),
                {'a': 'adesc',
                '__module__': 'svnmailer_test.unit.settings._typedstruct',
                'b': 'bdesc', '__slots__': ('foo', 'bar')},
            '_generateSpecials', '*private*',
            'instancemethod', None, Foo,
            'Foo', {'init': 'kw', '__cls': True},
            {'init': 'kw', '__cls': True}
        )

    def test08generateSpecials(self):
        """ ._generateSpecials returns a dict of special method functions """
        class MyCreator(self._mycreator):
            def _generateSetItem(this, private):
                return "setitem(%s)" % private
            def _generateGetItem(this, private):
                return "getitem(%s)" % private
            def _generateRepr(this, private):
                return "repr(%s)" % private
            def _generateCall(this, private):
                return "call(%s)" % private
            def _generateEq(this, private, iseq = True):
                return "eq(%s, %s)" % (private, iseq)
        inst = MyCreator()
        self.mock.result.append(inst._generateSpecials('*private*'))
        inst._eqignore = []
        self.mock.result.append(inst._generateSpecials('*priv2*'))

        self.expect({
            '__call__'   : 'call(*private*)',
            '__setitem__': 'setitem(*private*)',
            '__repr__'   : 'repr(*private*)',
            '__getitem__': 'getitem(*private*)'
        }, {
            '__call__'   : 'call(*priv2*)',
            '__setitem__': 'setitem(*priv2*)',
            '__repr__'   : 'repr(*priv2*)',
            '__getitem__': 'getitem(*priv2*)',
            '__eq__'     : 'eq(*priv2*, True)',
            '__ne__'     : 'eq(*priv2*, False)',
        })

    def test09generateMembers(self):
        """ ._generateMembers returns a dict of member descriptors """
        inst = self._mycreator2()
        maps = {'bar': 'zonk'}
        self.mock.result.append(
            inst._generateMembers('*private*', maps, '*arg*')
        )

        self.expect({
            'foo': 'descriptor(foo, *private*, member1(None, *arg*, None))',
            'bar': 'descriptor(bar, *private*, member2(zonk, *arg*, 1))'
        })

    def test10generateMembers(self):
        """ ._generateMembers applies the aliases """
        inst = self._mycreator2()
        inst._members[('foo', 'plopp')] = inst._members.pop(('foo',))
        maps = {'bar': 'zonk'}
        self.mock.result.append(
            inst._generateMembers('*private*', maps, '*arg*')
        )

        self.expect({
            'foo'  : 'descriptor(foo, *private*, member1(None, *arg*, None))',
            'bar'  : 'descriptor(bar, *private*, member2(zonk, *arg*, 1))',
            'plopp': 'descriptor(foo, *private*, member1(None, *arg*, None))',
        })

    def test11prepareMembers(self):
        """ ._prepareMembers prepares the members properly from dict """
        inst = self._mycreator2()
        members, names = inst._prepareMembers(
            {'a': 'foo', 'b': None, 'c': ('bar', {'some': 'param'})},
            {'foo': 'footype', 'bar': 'bartype', None: 'nonetype'}
        )
        self.failUnless(isinstance(names, tuple), "No tuple")
        names = list(names)
        names.sort()
        self.mock.result.extend([members, names])

        self.expect(
            {('a',): ('footype', None), ('b',): ('nonetype', None),
                ('c',): ('bartype', {'some': 'param'})},
            ['a', 'b', 'c']
        )

    def test12prepareMembers(self):
        """ ._prepareMembers prepares the members properly from tuple """
        inst = self._mycreator2()
        inst._DEFAULTMEMBER = '*defaultmember'
        members, names = inst._prepareMembers(('a', 'b', 'c'), {})
        self.failUnless(isinstance(names, tuple), "No tuple")
        names = list(names)
        names.sort()
        self.mock.result.extend([members, names])

        self.expect(
            {('a',): ('*defaultmember', None),
                ('b',): ('*defaultmember', None),
                ('c',): ('*defaultmember', None)},
            ['a', 'b', 'c']
        )

    def test13prepareMembers(self):
        """ ._prepareMembers rejects invalid member names """
        inst = self._mycreator2()

        self.failUnlessRaises(AssertionError,
            inst._prepareMembers, ('a', '__b'), {})

    def test14prepareMembers(self):
        """ ._prepareMembers rejects invalid member specs """
        inst = self._mycreator2()

        self.failUnlessRaises(AssertionError,
            inst._prepareMembers, {'a': self}, {})
        self.failUnlessRaises(AssertionError,
            inst._prepareMembers, {'a': ('foo', 'bar', 'baz')}, {})

    def test15prepareMembers(self):
        """ ._prepareMembers rejects unknown member specs """
        inst = self._mycreator2()

        self.failUnlessRaises(AssertionError,
            inst._prepareMembers, {'a': 'foo'}, {'bar': 'baz'})

    def test16prepareMembers(self):
        """ ._prepareMembers applies ._aliases properly """
        inst = self._mycreator2()
        inst._aliases = {'x': 'a', 'y': 'b'}
        members, names = inst._prepareMembers(('a', 'b'), {None: 'gnarz'})
        names = list(names)
        names.sort()
        self.mock.result.extend([members, names])

        self.expect(
            {('b', 'y'): ('gnarz', None), ('a', 'x'): ('gnarz', None)},
            ['a', 'b']
        )

    def test17prepareMembers(self):
        """ ._prepareMembers rejects invalid alias names """
        inst = self._mycreator2()
        inst._aliases = {'x': 'a', '__y': 'b'}

        self.failUnlessRaises(AssertionError,
            inst._prepareMembers, ('a', 'b'), {})

    def test18prepareMembers(self):
        """ ._prepareMembers rejects aliases pointing to unknown members """
        inst = self._mycreator2()
        inst._aliases = {'x': 'a', 'y': 'b'}

        self.failUnlessRaises(AssertionError,
            inst._prepareMembers, ('a', 'c'), {})

    def test19generateSetItem(self):
        """ ._generateSetItem generates a __setitem__ function """
        inst = self._mycreator2()
        private = base.Store(subst = {})
        func = inst._generateSetItem(private)
        func('x', 'key1', 'value1')
        func('x', 'key2', 'value2')
        self.mock.result.append(private.subst)

        self.expect({'key2': 'value2', 'key1': 'value1'})

    def test20generateGetItem(self):
        """ ._generateGetItem generates a __getitem__ function """
        inst = self._mycreator2()
        private = base.Store(subst = {'key1': 'v1', 'key2': 'v2'})
        func = inst._generateGetItem(private)
        self.mock.result.extend([func('x', 'key1'), func('x', 'key2')])

        self.expect('v1', 'v2')

    def test21generateRepr(self):
        """ ._generateRepr generates a __repr__ function """
        inst = self._mycreator2()
        inst._cls = base.Store(__module__ = 'foo.module', __name__ = 'bar')
        private = base.Store(values = {})
        func = inst._generateRepr(private)
        self.mock.result.append(func('x'))
        private.values['a'] = 10
        private.values['b'] = 11
        self.mock.result.append(func('x'))

        self.expect(
            'foo.module.bar()',
            'foo.module.bar(\n    a = 10,\n    b = 11\n)'
        )

    def test22generateCall(self):
        """ ._generateCall generates a __call__ function """
        inst = self._mycreator2()
        private = base.Store(
            members = '*members', subst = {'subst': 'record'},
            values = {'value': 'ofkey'}
        )
        func = inst._generateCall(private)
        self.mock.result.extend([
            func('x', 'members'), func('x', 'subst'), func('x', 'values')
        ])

        self.failUnless(isinstance(self.mock.result[-1],
            self.mock['svnmailer.util'].ReadOnlyDict), "Not readonly")
        self.failUnless(isinstance(self.mock.result[-2],
            self.mock['svnmailer.util'].ReadOnlyDict), "Not readonly")
        self.expect('*members', {'subst': 'record'}, {'value': 'ofkey'})

    def test23generateCall(self):
        """ ._generateCall's function raises KeyError """
        inst = self._mycreator2()
        private = base.Store()
        func = inst._generateCall(private)

        self.failUnlessRaises(KeyError, func, 'x', 'foo')

    def test24generateEq(self):
        """ ._generateEq generates a __eq__ function """
        inst = self._mycreator2()
        private = base.Store(members = ('a', 'b'), eqignore = {})
        func = inst._generateEq(private)
        this = base.Store(a = 10, b = 11)
        other = base.Store(a = 10, b = 11)
        self.mock.result.append(func(this, other))
        this.a = 12
        self.mock.result.append(func(this, other))
        other.a = 12
        this.b = 13
        self.mock.result.append(func(this, other))

        self.expect(True, False, False)

    def test25generateEq(self):
        """ ._generateEq's function deals with missing attributes """
        inst = self._mycreator2()
        private = base.Store(members = ('a', 'b'), eqignore = {})
        func = inst._generateEq(private)
        this = base.Store(a = 10, b = 11)
        other = base.Store(a = 10)
        self.mock.result.append(func(this, other))

        self.expect(False)

    def test26generateEq(self):
        """ ._generateEq's function deals with different types """
        inst = self._mycreator2()
        private = base.Store(members = ('a', 'b'), eqignore = {})
        func = inst._generateEq(private)
        this = base.Store(a = 10, b = 11)
        other = 1
        self.mock.result.append(func(this, other))

        self.expect(False)

    def test27generateEq(self):
        """ ._generateEq's function deals with private.eqignore properly """
        inst = self._mycreator2()
        private = base.Store(members = ('a', 'b'), eqignore = {'a': None})
        func = inst._generateEq(private)
        this = base.Store(a = 10, b = 11)
        other = base.Store(a = 10, b = 11)
        self.mock.result.append(func(this, other))
        this.a = 12
        self.mock.result.append(func(this, other))
        this.b = 13
        self.mock.result.append(func(this, other))

        self.expect(True, True, False)

    def test28generateEq(self):
        """ ._generateEq generates a __ne__ function """
        inst = self._mycreator2()
        private = base.Store(members = ('a', 'b'), eqignore = ())
        func = inst._generateEq(private, False)
        this = base.Store(a = 10, b = 11)
        other = base.Store(a = 10, b = 11)
        self.mock.result.append(func(this, other))
        this.a = 12
        self.mock.result.append(func(this, other))
        other.a = 12
        this.b = 13
        self.mock.result.append(func(this, other))

        self.expect(False, True, True)

    def test29createPrivate(self):
        """ ._createPrivate returns a Private instance """
        inst = self._mycreator()
        inst._eqignore = {1: 2}
        class Private(object):
            def __new__(cls, names, eqignore):
                self.mock.result.extend(['Private', names, eqignore])
                return "*private*"
        self._struct.Private = Private
        self.mock.result.append(inst._createPrivate())

        self.expect('Private', ('foo', 'bar'), {1: 2}, '*private*')

    def test30createDescriptor(self):
        """ ._createDescriptor returns a Descriptor instance """
        inst = self._mycreator()
        class Descriptor(object):
            def __new__(cls, name, private, member):
                self.mock.result.extend(['Descriptor', name, private, member])
                return '*descriptor*'
        self._struct.Descriptor = Descriptor
        self.mock.result.append(
            inst._createDescriptor('*name', '*private*', '*member')
        )

        self.expect(
            'Descriptor', '*name', '*private*', '*member', '*descriptor*'
        )

    def test31createMetaClass(self):
        """ ._createMetaClass returns a new MetaClass instance """
        inst = self._mycreator()
        class MetaClass(object):
            def __new__(cls, name, bases, cdict):
                self.mock.result.extend(['MetaClass', name, bases, cdict])
                return '*metaclass*'
        self._struct.MetaClass = MetaClass
        self.mock.result.append(
            inst._createMetaClass('*name*', '*bases*', '*cdict*')
        )

        self.expect('MetaClass', '*name*', '*bases*', '*cdict*', '*metaclass*')

_tests.append(StructCreatorTestCase)


class PrivateTestCase(TestBase):
    """ svnmailer.settings._typedstruct.Private """
    PREFIX = "Private"

    def test00general(self):
        """ Private is a private object """
        expected = object
        result = self._struct.Private

        self.failUnless(issubclass(result, expected),
            "Is not an %r, but %r" % (expected, result))
        self.failIf(result.__name__ in self.public(self._struct),
            "Is public")

    def test01init(self):
        """ .__init__ initializes instance attributes properly """
        inst = self._struct.Private('*names*', '*eqignore*')
        self.mock.result = [
            inst.members, inst.eqignore, inst.values, inst.subst
        ]

        self.failIf(self.mock.result[-1] is self.mock.result[-2],
            "Equal dicts")
        self.expect('*names*', '*eqignore*', {}, {})

_tests.append(PrivateTestCase)


class DescriptorTestCase(TestBase):
    """ svnmailer.settings._typedstruct.Descriptor """
    PREFIX = "Descriptor"

    def setUp(self):
        """ Extends the test environment """
        super(DescriptorTestCase, self).setUp()
        class MyMember(object):
            def __init__(this):
                this.mapper = '*12345*'
                this.instance = None
            def premap(this, value):
                self.mock.result.extend(['premap', value, this.instance])
                return 'premapped(%s)' % value
            def transform(this, value):
                self.mock.result.extend(['transform', value, this.instance])
                return 'transformed(%s)' % value
            def substitute(this, value, subst):
                self.failUnless(isinstance(subst,
                    self.mock['svnmailer.util'].ReadOnlyDict), "Not readonly")
                self.mock.result.extend(['substitute',
                    value, subst, this.instance
                ])
                return 'substituted(%s, %s)' % (value, subst)
            def postmap(this, value):
                self.mock.result.extend(['postmap', value, this.instance])
                return 'postmapped(%s)' % value

        class MyDescriptor(self._struct.Descriptor):
            def __init__(this):
                this.name = 'foo'
                this.private = base.Store(values = {}, subst = {})
                this.member = MyMember()

        self._mydescriptor = MyDescriptor

    def test00general(self):
        """ Descriptor is a private object """
        expected = object
        result = self._struct.Descriptor

        self.failUnless(issubclass(result, expected),
            "Is not an %r, but %r" % (expected, result))
        self.failIf(result.__name__ in self.public(self._struct),
            "Is public")

    def test01init(self):
        """ .__init__ initializes instance attributes properly """
        inst = self._struct.Descriptor('*name*', '*private*', '*member*')
        self.mock.result = [inst.name, inst.private, inst.member]

        self.expect('*name*', '*private*', '*member*')

    def test02set(self):
        """ .__set__ premaps and transforms the value; instance is set """
        inst = self._mydescriptor()
        inst.__set__('*instance', 'zonk')
        self.mock.result.extend([inst.member.instance, inst.private.values])

        self.expect(
            'premap', 'zonk', '*instance',
            'transform', 'premapped(zonk)', '*instance',
            None, {'foo': 'transformed(premapped(zonk))'}
        )

    def test03set(self):
        """ .__set__ doesn't premap with None mapper """
        inst = self._mydescriptor()
        inst.member.mapper = None
        inst.__set__('*instance2', 'buh')
        self.mock.result.extend([inst.member.instance, inst.private.values])

        self.expect(
            'transform', 'buh', '*instance2',
            None, {'foo': 'transformed(buh)'}
        )

    def test04get(self):
        """ .__get__ substitutes and postmaps the value; instance is set """
        inst = self._mydescriptor()
        inst.private.values['foo'] = 'plopp'
        inst.private.subst['x'] = 'y'
        res = inst.__get__('*instance', '*owner')
        self.mock.result.extend([inst.member.instance, res])

        self.expect(
            'substitute', 'plopp', {'x': 'y'}, '*instance',
            'postmap', "substituted(plopp, {'x': 'y'})", '*instance',
            None, "postmapped(substituted(plopp, {'x': 'y'}))"
        )

    def test05get(self):
        """ .__get__ doesn't postmap with None mapper """
        inst = self._mydescriptor()
        inst.private.values['foo'] = 'plopp'
        inst.private.subst['x'] = 'y'
        inst.member.mapper = None
        res = inst.__get__('*instance', '*owner')
        self.mock.result.extend([inst.member.instance, res])

        self.expect(
            'substitute', 'plopp', {'x': 'y'}, '*instance',
            None, "substituted(plopp, {'x': 'y'})"
        )

    def test06get(self):
        """ .__get__ returns None on None instance """
        inst = self._mydescriptor()
        self.mock.result.append(inst.__get__(None, '*owner'))

        self.expect(None)

    def test07delete(self):
        """ .__delete__ removes the value entirely """
        inst = self._mydescriptor()
        inst.private.values['foo'] = 'bar'
        inst.__delete__('*instance')
        self.mock.result.append(inst.private.values)

        self.expect({})

    def test08delete(self):
        """ .__delete__ catches KeyError """
        inst = self._mydescriptor()
        inst.private.values['bar'] = 'baz'
        inst.__delete__('*instance2')
        self.mock.result.append(inst.private.values)

        self.expect({'bar': 'baz'})

_tests.append(DescriptorTestCase)


#### Glue code                                      ####
########################################################
__all__.extend(_tests)

def getSuite():
    """ Returns the test suite for this module """
    return [_test.__name__ for _test in _tests]
